/*****************************************************************************/
/*                      WiNRADiO G69DDC EXCELSIOR ULTRA                      */
/*                                SDK DEMO                                   */
/*                   Interlayer between G69DDC API and GUI                   */
/*                        (c) Copyright RADIXON 2022                         */
/*****************************************************************************/

#ifndef __U_RECEIVER_H__
#define __U_RECEIVER_H__

#include <wx/wx.h>
#include <wx/thread.h>
#include "g69ddcapi.h"
#include "u_defs.h"
#include "u_types.h"
#include "u_fft.h"

//minimam DDC1/DDC2 BW which allows FMW mode
#define MIN_FMW_DDC2_BW             100000

//number of supported demodulator modes
#define DEM_MODE_COUNT              9

#define SWEEPER_STEP_COUNT          4

//AGC modes
#define AGC_OFF                     0
#define AGC_SLOW                    1
#define AGC_MEDIUM                  2
#define AGC_FAST                    3

//MUTE_BITS_MANUAL - mute bit, set when manual mute (Mute button in GUI)
#define MUTE_BITS_MANUAL            0x00000001
//MUTE_BITS_INACTIVE_CHANNEL - mute bit, set when given channel is inactive (RX1/RX2/RX3 buttons in GUI)
#define MUTE_BITS_INACTIVE_CHANNEL  0x00000002
//MUTE_BITS_SQUELCH - mute bit, set when squelched
#define MUTE_BITS_SQUELCH           0x00000004

//receiver event codes
typedef enum
{
    ReceiverEvent_Power=0,              //power state changed
    ReceiverEvent_Attenuator,           //attenuator changed
    ReceiverEvent_Preamplifier,         //preamplifier state changed
    ReceiverEvent_MWFilter,             //MW filter changed
    ReceiverEvent_IFRBWIndex,           //IF spectrum RBW changed
    ReceiverEvent_ActiveChannel,        //active channel (RX1/RX2) changed
    ReceiverEvent_SweeperStarted,       //sweeper started
    ReceiverEvent_SweeperStopped,       //sweeper stopped
    ReceiverEvent_FrontEndFrequency,    //front-end frequency changed
    ReceiverEvent_Range,                //receiver's input range changed
    ReceiverEvent_DDC1,                 //DDC1 BW (DDC type) changed
    ReceiverEvent_DDC1_RBWList,         //DDC1 spectrum RBW list changed
    ReceiverEvent_DDC1_RBWIndex,        //DDC1 spectrum RBW changed
    ReceiverEvent_DDC1_Frequency,       //DDC1 relative frequency changed
    ReceiverEvent_DDC1_AbsFrequency,    //DDC1 absolute frequency changed
    ReceiverEvent_DDC2_RBWList          //DDC2 spectrum RBW list changed
} TReceiverEvent;

//channel related event codes
typedef enum
{
    ChannelEvent_DDC2_Count,            //number of available DDC2 types changed
    ChannelEvent_DDC2,                  //DDC2 BW (DDC type) changed
    ChannelEvent_DDC2_RBWList,          //DDC2 spectrum RBW list changed
    ChannelEvent_DDC2_RBWIndex,         //DDC2 spectrum RBW changed
    ChannelEvent_DDC2_Frequency,        //DDC2 relative frequency changed
    ChannelEvent_DDC2_AbsFrequency,     //DDC2 absolute frequency changed

    ChannelEvent_DEM_Frequency,         //demod relative frequency changed
    ChannelEvent_DEM_AbsFrequency,      //demod absolute frequency changed
    ChannelEvent_DEM_Bandwidth,         //demod filter bandwidth changed
    ChannelEvent_DEM_Mode,              //demod mode changed
    ChannelEvent_DEM_AMSSideband,       //AMS side band changed
    ChannelEvent_DEM_DSBSideband,       //DSB side band changed
    ChannelEvent_DEM_ISBSideband,       //ISB side band changed
    ChannelEvent_DEM_CWTone,            //CW tone frequency changed
    ChannelEvent_DEM_FMWStereo,         //FMW stereo decoder enabled/disabled
    ChannelEvent_DEM_FMWDeemphasis,     //FMW de-emphasis changed

    ChannelEvent_Audio,                 //audio sample rate (ASR) changed
    ChannelEvent_Audio_RBWList,         //audio spectrum RBW list changed
    ChannelEvent_Audio_RBWIndex,        //audio spectrum RBW changed
    ChannelEvent_Audio_Volume,          //volume changed
    ChannelEvent_Audio_Mute,            //mute/unmode
    ChannelEvent_Audio_Squelch,         //squelch state changed
    ChannelEvent_Audio_Filter,          //audio filter enabled/disabled
    ChannelEvent_Audio_FilterParams,    //audio filter params changed

    ChannelEvent_AGC,                   //AGC changed
    ChannelEvent_Gain                   //gain changed

} TChannelEvent;

typedef enum
{
    Range_1,
    Range_2
} TRange;

//callbacks, GUI implements the following methods to receive events, frequency spectrums etc. from TReceiver object
class TReceiverCallback
{
public:
    virtual void ReceiverCallback_Event(TReceiverEvent Event)=0;
    virtual void ReceiverCallback_ChannelEvent(uint32_t Channel,TChannelEvent Event)=0;
    virtual void ReceiverCallback_DDC1Spectrum(float *Data,uint32_t Count,double Range,double Offset)=0;
    virtual void ReceiverCallback_DDC2Spectrum(uint32_t Channel,float *Data,uint32_t Count,double Range,double Offset)=0;
    virtual void ReceiverCallback_AudioSpectrum(uint32_t Channel,float *Data,uint32_t Count,uint32_t Range)=0;
    virtual void ReceiverCallback_IFSpectrum(float *Data,uint32_t Count,double Range,double Offset)=0;
    virtual void ReceiverCallback_Audio(uint32_t Channel,const float *Samples,uint32_t Count)=0;
    virtual void ReceiverCallbask_Sweeper(float *Data,uint32_t Count)=0;
    virtual bool ReceiverCallback_SweeperQueryEnd(void)=0;
};

//channel data object
class TChannel
{
public:
    uint32_t            m_DDC2_ChannelId;
    uint32_t            m_DDC2_Count;
    uint32_t            m_DDC2_Index;
    G69DDC_DDC_INFO     m_DDC2_Info;
    int32_t             m_DDC2_Frequency;   //relative DDC2 center frequency
    uint64_t            m_DDC2_AbsFrequency;//absolute DDC2 center frequency
    uint32_t            m_DDC2_RBW_Index;   //index of current DDC2 spectrum RBW
    uint32_t            m_DDC2_FFT_Size;    //DDC2 FFT size
    float               m_DDC2_FFT_Buffer[MAX_FFT_SIZE*2];//buffer used by DDC2 FFT
    float               m_DDC2_FFT_Coeffs[MAX_FFT_SIZE];//DDC2 FFT coeffs (normalization coeffs * window coeffs)
    TFFTBuffer          m_DDC2_FFT_OverlappingBuffer;//DDC2 FFT overlapping buffer
    float               m_DDC2_SpectrumCompensation[MAX_FFT_SIZE]; //DDC2 spectrum compensation data
    wxMutex             m_DDC2_FFT_Mutex;
    double              m_DDC2_FFT_Range;   //total frequency range of DDC2 frequency spectrum (DDC1RBW*(LastBin-FirstBin+1)) sent to GUI
    double              m_DDC2_FFT_Offset;  //offset of start of useful DDC1 band in total frequency range
    uint32_t            m_DDC2_FFT_FirstBin;//the first DDC2 FFT bin where useful band starts
    uint32_t            m_DDC2_FFT_BinCount;//the last DDC2 FFT bin where useful band ends

    int32_t             m_DEM_Frequency;    //demod relative frequency
    uint64_t            m_DEM_AbsFrequency; //demod absolute frequency (displayed in frequency edit box, below RX1/RX2/RX3 buttons)
    uint32_t            m_DEM_Bandwidth;    //demod bandwidth
    uint32_t            m_DEM_Mode;         //current demod mode
    uint32_t            m_DEM_AMSSideband;  //AMS side band
    uint32_t            m_DEM_DSBSideband;  //DSB side band
    uint32_t            m_DEM_ISBSideband;  //ISB side band
    uint32_t            m_DEM_CWTone;       //CW tone frequency
    BOOL                m_DEM_FMWStereo;    //FMW stereo
    uint32_t            m_DEM_FMWDeemphasis;//FMW de-emphasis (micro-seconds)


    uint32_t            m_Audio_Index;      //index of current audio sample rate
    uint32_t            m_Audio_SampleRate; //current audio sample rate
    uint32_t            m_Audio_RBW_Index;  //index of current audio spectrum RBW
    wxMutex             m_Audio_FFT_Mutex;
    uint32_t            m_Audio_FFT_Size;   //audio spectrum FFT size
    float               m_Audio_FFT_Buffer[MAX_FFT_SIZE*2]; //buffer used by audio FFT
    float               m_Audio_FFT_Coeffs[MAX_FFT_SIZE]; //audio FFT coeffs (normalization coeffs * window coeffs)
    TFFTBuffer          m_Audio_FFT_OverlappingBuffer; //audio spectrum FFT overlapping buffer
    uint32_t            m_Audio_SamplesPerBuffer; //number of sample sets be audio stream buffer

    bool                m_Audio_Filter;             //audio filter enabled/disabled
    uint32_t            m_Audio_Filter_CutOffLow;   //audio filter cut-off low frequency
    uint32_t            m_Audio_Filter_CutOffHigh;  //audio filter cut-off high frequency
    double              m_Audio_Filter_Deemphasis;  //audio filter deemphasis
    uint32_t            m_Audio_Volume;
    float               m_Audio_VolumeCoeff;

    uint32_t            m_Audio_MuteBits;


    uint32_t            m_AGC;          //AGC
    int32_t             m_Gain;         //gain

    inline TChannel(void):m_DDC2_FFT_OverlappingBuffer(MAX_FFT_SIZE),m_DDC2_FFT_Mutex(wxMUTEX_RECURSIVE),
        m_Audio_FFT_Mutex(wxMUTEX_RECURSIVE),m_Audio_FFT_OverlappingBuffer(MAX_FFT_SIZE)
    {
    }
};

class TReceiver:public IG69DDCDeviceCallback
{
public:
    inline bool GetSweeper(void) const
    {
        return m_Sweeper;
    }

    bool StartSweeper(int64_t Start,int64_t End,uint32_t RBWIndex);
    void StopSweeper(void);
    void GetSweeperParams(int64_t *Start,int64_t *End,double *RBW);

    inline uint64_t GetFrontEndFrequency(void)
    {
        return m_FrontEndFrequency;
    }

    void SetFrontEndFrequency(uint64_t FrontEndFrequency);

    inline TRange GetRange(void) const
    {
        return m_Range;
    }

    inline uint32_t GetChannelCount(void) const
    {
        return m_ChannelCount;
    }

    inline uint32_t GetDemModeCount(void) const
    {
        return DEM_MODE_COUNT; //except DRM mode
    }

    wxString GetDemModeName(uint32_t Mode);

    void SetDemMode(uint32_t ChannelIndex,uint32_t ModeIndex);

    inline uint32_t GetDemMode(uint32_t ChannelIndex,uint32_t *ModeCode=NULL) const
    {
        if(ModeCode)
        {
            *ModeCode=m_ModeCodes[m_Channel[ChannelIndex].m_DEM_Mode];
        }

        return m_Channel[ChannelIndex].m_DEM_Mode;
    }

    inline float GetSignalLevel(uint32_t ChannelIndex)
    {
     float RMS;

        m_Device->GetSignalLevel(m_Channel[ChannelIndex].m_DDC2_ChannelId,NULL,&RMS);

        return RMS;
    }

    inline void GetDDCInfo(uint32_t TypeIndex,G69DDC_DDC_INFO *Info)
    {
        m_Device->GetDDCInfo(TypeIndex,Info);
    }

    inline uint32_t GetDDC1Count(void) const
    {
        return m_DDC1_Count;
    }

    void SetDDC1(uint32_t TypeIndex);

    inline uint32_t GetDDC1(void) const
    {
        return m_DDC1_Index;
    }
    inline G69DDC_DDC_INFO GetDDC1Info(void) const
    {
        return m_DDC1_Info;
    }

    void GetDDC1RBWs(double *RBWs);

    inline uint32_t GetDDC1RBWIndex(void) const
    {
        return m_DDC1_RBW_Index;
    }
    void SetDDC1RBWIndex(uint32_t RBWIndex);


    inline int32_t GetDDC1Frequency(void) const
    {
        return m_DDC1_Frequency;
    }

    inline uint64_t GetDDC1AbsFrequency(void) const
    {
        return m_DDC1_AbsFrequency;
    }

    void SetDDC1AbsFrequency(uint64_t Frequency);

    inline uint32_t GetDDC2Count(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_Count;
    }

    void SetDDC2(uint32_t ChannelIndex,uint32_t TypeIndex);

    inline uint32_t GetDDC2(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_Index;
    }

    inline G69DDC_DDC_INFO GetDDC2Info(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_Info;
    }

    void GetDDC2RBWs(uint32_t ChannelIndex,double *RBWs);

    inline uint32_t GetDDC2RBWIndex(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_RBW_Index;
    }

    void SetDDC2RBWIndex(uint32_t ChannelIndex,uint32_t RBWIndex);

    inline int32_t GetDDC2Frequency(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_Frequency;
    }

    void SetDDC2Frequency(uint32_t ChannelIndex,int32_t Frequency);

    inline uint64_t GetDDC2AbsFrequency(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DDC2_AbsFrequency;
    }

    inline int32_t GetDemFrequency(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_Frequency;
    }

    void SetDemFrequency(uint32_t ChannelIndex,int32_t Frequency);

    inline uint64_t GetDemAbsFrequency(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_AbsFrequency;
    }

    void SetDemAbsFrequency(uint32_t ChannelIndex,uint64_t Frequency);

    inline uint32_t GetDemBandwidth(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_Bandwidth;
    }

    void SetDemBandwidth(uint32_t ChannelIndex,uint32_t Bandwidth);

    inline uint32_t GetDemAMSSideband(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_AMSSideband;
    }

    void SetDemAMSSideband(uint32_t ChannelIndex,uint32_t Sideband);

    inline uint32_t GetDemDSBSideband(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_DSBSideband;
    }

    void SetDemDSBSideband(uint32_t ChannelIndex,uint32_t Sideband);

    inline uint32_t GetDemISBSideband(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_ISBSideband;
    }

    void SetDemISBSideband(uint32_t ChannelIndex,uint32_t Sideband);

    inline uint32_t GetDemCWTone(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_CWTone;
    }

    void SetDemCWTone(uint32_t ChannelIndex,uint32_t Frequency);

    inline bool GetDemFMWStereo(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_FMWStereo!=FALSE;
    }

    void SetDemFMWStereo(uint32_t ChannelIndex,bool Stereo);

    inline uint32_t GetDemFMWDeemphasis(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_DEM_FMWDeemphasis;
    }

    void SetDemFMWDeemphasis(uint32_t ChannelIndex,uint32_t Deemphasis);

    inline int32_t GetDemTuneError(uint32_t ChannelIndex)
    {
     int32_t Error;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_TUNE_ERROR,&Error,sizeof(Error));

        return Error;
    }

    inline double GetAMDepth(uint32_t ChannelIndex)
    {
     double AMDepth;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_AM_DEPTH,&AMDepth,sizeof(AMDepth));

        return AMDepth;
    }

    inline uint32_t GetDeviation(uint32_t ChannelIndex)
    {
     uint32_t Deviation;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_FM_DEVIATION,&Deviation,sizeof(Deviation));

        return Deviation;
    }

    inline bool IsDemAMSLocked(uint32_t ChannelIndex)
    {
     BOOL Locked;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_AMS_LOCK,&Locked,sizeof(Locked));

        return Locked!=FALSE;
    }

    inline bool IsDemDSBLocked(uint32_t ChannelIndex)
    {
     BOOL Locked;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_DSB_LOCK,&Locked,sizeof(Locked));

        return Locked!=FALSE;
    }

    inline bool IsDemFMWStereo(uint32_t ChannelIndex)
    {
     BOOL Stereo;

        m_Device->GetDemodulatorState(m_Channel[ChannelIndex].m_DDC2_ChannelId,G69DDC_DEMODULATOR_STATE_FMW_STEREO,&Stereo,sizeof(Stereo));

        return Stereo!=FALSE;
    }

    inline double GetCurrentGain(uint32_t ChannelIndex)
    {
     double Gain;

        m_Device->GetCurrentGain(m_Channel[ChannelIndex].m_DDC2_ChannelId,&Gain);

        return Gain;
    }

    inline bool GetMute(uint32_t ChannelIndex) const
    {
        return (m_Channel[ChannelIndex].m_Audio_MuteBits & MUTE_BITS_MANUAL)!=0;
    }

    void SetMute(uint32_t ChannelIndex,bool Mute);

    inline bool GetSquelched(uint32_t ChannelIndex) const
    {
        return (m_Channel[ChannelIndex].m_Audio_MuteBits & MUTE_BITS_SQUELCH)!=0;
    }

    void SetSquelched(uint32_t ChannelIndex,bool Squelched);

    uint8_t GetVolume(uint32_t ChannelIndex);

    void SetVolume(uint32_t ChannelIndex,uint8_t Volume);

    inline uint32_t GetAudioSampleRateCount(void) const
    {
        uint32_t Count;

        m_Device->GetAudioSampleRateCount(&Count);

        return Count;
    }

    inline uint32_t GetAudioSampleRate(uint32_t Index) const
    {
        uint32_t SampleRate;

        m_Device->GetAudioSampleRate(Index,&SampleRate);

        return SampleRate;
    }

    inline uint32_t GetAudio(uint32_t ChannelIndex,uint32_t *SampleRate) const
    {
        if(SampleRate)
        {
            *SampleRate=m_Channel[ChannelIndex].m_Audio_SampleRate;
        }

        return m_Channel[ChannelIndex].m_Audio_Index;
    }

    void SetAudio(uint32_t ChannelIndex,uint32_t Index);

    void GetAudioRBWs(uint32_t ChannelIndex,double *RBWs);

    inline uint32_t GetAudioRBWIndex(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_Audio_RBW_Index;
    }
    void SetAudioRBWIndex(uint32_t ChannelIndex,uint32_t RBWIndex);

    inline bool GetAudioFilter(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_Audio_Filter;
    }

    void SetAudioFilter(uint32_t ChannelIndex,bool Enabled);

    inline void GetAudioFilterParams(uint32_t ChannelIndex,uint32_t *Low,uint32_t *High,double *Deem)
    {
        *Low=m_Channel[ChannelIndex].m_Audio_Filter_CutOffLow;
        *High=m_Channel[ChannelIndex].m_Audio_Filter_CutOffHigh;
        *Deem=m_Channel[ChannelIndex].m_Audio_Filter_Deemphasis;
    }

    void SetAudioFilterParams(uint32_t ChannelIndex,uint32_t Low,uint32_t High,double Deem);

    inline uint32_t GetAGC(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_AGC;
    }

    void SetAGC(uint32_t ChannelIndex,uint32_t AGC);

    inline int32_t GetGain(uint32_t ChannelIndex) const
    {
        return m_Channel[ChannelIndex].m_Gain;
    }

    void SetGain(uint32_t ChannelIndex,int32_t Gain);

    inline uint32_t GetActiveChannel(void) const
    {
        return m_ActiveChannel;
    }
    void SetActiveChannel(uint32_t ChannelIndex);

    inline uint16_t GetADCLevel(void) const
    {
        return m_ADCLevel;
    }


    inline uint32_t GetIFRBWIndex(void) const
    {
        return m_IF_RBW_Index;
    }

    void SetIFRBWIndex(uint32_t Index);

    void GetIFRBWs(double *RBWs); //RBWs is pointer to array with RBW_COUNT elements

    void FrontEndFrequencyUp(void);
    void FrontEndFrequencyDown(void);

    void SetPreamplifier(bool Preamp);
    inline bool GetPreamplifier(void) const
    {
        return m_Preamp;
    }

    void SetMWFilter(bool MWFilter);
    inline bool GetMWFilter(void) const
    {
        return m_MWFilter;
    }

    void SetAttenuator(uint32_t Attenuator);
    inline uint32_t GetAttenuator(void) const
    {
        return m_Attenuator;
    }

    void SetPower(bool Power);
    inline bool GetPower(void) const
    {
        return m_Power;
    }

    inline bool IsConnected(void)
    {
        BOOL Connected;

        m_Device->IsConnected(&Connected);

        return Connected!=FALSE;
    }

    inline bool GetDeviceState(G69DDC_DEVICE_STATE *State)
    {
        return m_Device->GetDeviceState(State)!=FALSE;
    }

    inline void GetDeviceInfo(G69DDC_DEVICE_INFO &Info) const
    {
        Info=m_DeviceInfo;
    }

    inline const G69DDC_DEVICE_INFO& GetDeviceInfo(void) const
    {
        return m_DeviceInfo;
    }

    inline void SetCallback(TReceiverCallback *Callback)
    {
        m_Callback=Callback;
    }

    bool Open(const char *DevicePath);
    void Close(void);

    TReceiver(void);
    virtual ~TReceiver(void);
private:
    IG69DDCDevice           *m_Device;          //G69DDC device object interface
    TReceiverCallback       *m_Callback;        //callback
    int                     m_Ignore;           //when non-zero, all the set methods are ignored
    G69DDC_DEVICE_INFO      m_DeviceInfo;       //device info
    bool                    m_Power;            //power state

    uint32_t                m_Attenuator;       //attenuator
    bool                    m_Preamp;           //preamplifier
    bool                    m_MWFilter;         //MW filter
    TChannel                *m_Channel;         //channels data
    uint32_t                m_ActiveChannel;    //index of active channel (RX1/RX2/RX3)
    uint64_t                m_FrontEndFrequency;//front-end frequency
    TRange                  m_Range;            //active receiver's input range

    uint32_t                m_IF_SamplesPerSnapshot;
    uint32_t                m_IF_BW;
    uint32_t                m_IF_RBW_Index;     //index of IF spectrum RBW
    uint32_t                m_IF_FFT_Size;      //IF spectrum FFT size
    float                   m_IF_FFT_Buffer[MAX_IF_FFT_SIZE*2]; //buffer used by IF spectrum FFT
    float                   m_IF_FFT_Coeffs[MAX_IF_FFT_SIZE]; //IF spectrum FFT coeffs (normalization coeffs * window coeffs * 1/32768)
    bool                    m_IF_UpdateSpectrumCompensation;
    float                   m_IF_SpectrumCompensation[MAX_IF_FFT_SIZE/2]; //IF spectrum compensation data
    wxMutex                 m_IF_FFT_Mutex;

    uint16_t                m_ADCLevel;             //the latest ADC level (0..32767), 32767 = ADC clipping

    uint32_t                m_ChannelCount;         //number of DDC2 channels

    uint32_t                m_DDC1_ChannelId;
    uint32_t                m_DDC1_Index;           //current DDC type index of DDC1
    G69DDC_DDC_INFO         m_DDC1_Info;            //current DDC type info of DDC1
    uint32_t                m_DDC1_Count;           //number of available DDC types of DDC1
    int32_t                 m_DDC1_Frequency;       //DDC1 center frequency
    uint64_t                m_DDC1_AbsFrequency;
    uint32_t                m_DDC1_RBW_Index;       //index of current DDC1 spectrum RBW
    uint32_t                m_DDC1_FFT_Size;        //DDC1 FFT size
    float                   m_DDC1_FFT_Buffer[MAX_FFT_SIZE*2];  //buffer used by DDC1 FFT
    float                   m_DDC1_FFT_Coeffs[MAX_FFT_SIZE];    //DDC1 FFT coeffs (normalization coeffs * window coeffs)
    TFFTBuffer              m_DDC1_FFT_OverlappingBuffer;       //DDC1 FFT overlapping buffer
    float                   m_DDC1_SpectrumCompensation[MAX_FFT_SIZE]; //DDC1 spectrum compensation data
    wxMutex                 m_DDC1_FFT_Mutex;
    double                  m_DDC1_FFT_Range;       //total frequency range of DDC1 frequency spectrum (DDC1RBW*(LastBin-FirstBin+1)) sent to GUI
    double                  m_DDC1_FFT_Offset;      //offset of start of useful DDC1 band in total frequency range
    uint32_t                m_DDC1_FFT_FirstBin;    //the first DDC1 FFT bin where useful band starts
    uint32_t                m_DDC1_FFT_BinCount;    //the last DDC1 FFT bin where useful band ends
    uint32_t                m_DDC1_SamplesPerBuffer;//number of sample sets per DDC1 stream buffer

    uint32_t                *m_DDC2_LUT;            //look up table to convert DDC2 channel ID to channel index

    bool                    m_Sweeper;                  //sweeper state
    int64_t                 m_Sweeper_StartFrequency;   //sweeper start frequency
    int64_t                 m_Sweeper_EndFrequency;     //sweeper end frequency
    uint32_t                m_Sweeper_FFT_Size;         //sweeper FFT size
    double                  m_Sweeper_FFT_RBW;          //sweeper RBW
    float                   m_Sweeper_FFT_Buffer[MAX_IF_FFT_SIZE*2];        //buffer for sweeper FFT
    float                   m_Sweeper_FFT_Coeffs[MAX_IF_FFT_SIZE];          //sweeper coeffs (normalization coeffs * window coeffs * 1/32768)
    float                   m_Sweeper_SpectrumCompensation[MAX_IF_FFT_SIZE/2]; //sweeper IF spectrum compensation data
    int64_t                 m_Sweeper_StartFrontEndFrequency;               //the first front-end window frequency which includes sweeper start frequency
    int64_t                 m_Sweeper_EndFrontEndFrequency;                //the last front-end window frequency which includes sweeper end frequency
    double                  m_Sweeper_Frequency;         //absolute end frequency of already processed part


    static const wxChar     *m_ModeNames[DEM_MODE_COUNT];
    static uint32_t         m_ModeCodes[DEM_MODE_COUNT];

    virtual void G69DDC_ADCSnapshotCallback(IG69DDCDevice *Device,const short *Buffer,uint32_t Count,uint32_t CenterFrequency,uint16_t ADCLevel);
    virtual void G69DDC_DDC1StreamCallback(IG69DDCDevice *Device,uint32_t DDC1ChannelId,const void *Buffer,uint32_t Count,uint32_t BitsPerSample);
    virtual int G69DDC_DDC1PlaybackStreamCallback(IG69DDCDevice *Device,uint32_t DDC1ChannelId,void *Buffer,uint32_t Count,uint32_t BitsPerSample);
    virtual void G69DDC_DDC2StreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,const float *Buffer,uint32_t Count);
    virtual void G69DDC_DDC2PreprocessedStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,const float *Buffer,uint32_t Count,float SlevelPeak,float SlevelRMS);
    virtual void G69DDC_AudioStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,uint32_t Stage,const float *Buffer,uint32_t Count);
    virtual int G69DDC_AudioPlaybackStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,float *Buffer,uint32_t Count);
    virtual int G69DDC_SweeperCallback(IG69DDCDevice *Device,const short *Buffer,uint32_t Count,uint32_t CenterFrequency,uint64_t FrontEndFrequency,uint32_t Repeat);

    void InvokeEvent(TReceiverEvent Event);
    void InvokeEvent(uint32_t Channel,TChannelEvent Event);

    void OnPowerChanged(void);
    void OnAttenuatorChanged(void);
    void OnPreaplifierChanged(void);
    void OnFrontEndFrequencyChanged(void);
    void OnIFRBWChanged(void);
    void UpdateIFFFTCoeffs(void);
    void UpdateSpectrumCompensationData(void);
    void UpdateIFSpectrumCompensationData(void);

    void OnDDC1Changed(void);
    void OnDDC1FrequencyChanged(bool UpdateHw);
    void UpdateDDC1SpectrumCompensationData(void);
    void UpdateDDC1FFTCoeffs(void);
    void StopDDC1(void);
    void GetDDC1FrequencyMinMax(int32_t *Min,int32_t *Max);

    void OnDDC2Changed(uint32_t ChannelIndex);
    void OnDDC2FrequencyChanged(uint32_t ChannelIndex);
    void OnDDC2AbsFrequencyChanged(uint32_t ChannelIndex);
    void UpdateDDC2SpectrumCompensationData(uint32_t ChannelIndex);
    void UpdateDDC2FFTCoeffs(uint32_t ChannelIndex);
    void StopDDC2(uint32_t ChannelIndex);

    void GetDemBandwidthMinMax(uint32_t ChannelIndex,uint32_t *Min,uint32_t *Max);
    void GetDemFrequencyMinMax(uint32_t ChannelIndex,int32_t *Min,int32_t *Max);
    void OnDemFrequencyChanged(uint32_t ChannelIndex);
    void OnDemAbsFrequencyChanged(uint32_t ChannelIndex);
    void OnDemBandwidthChanged(uint32_t ChannelIndex);
    void OnDemModeChanged(uint32_t ChannelIndex);

    void OnAudioChanged(uint32_t ChannelIndex);
    void UpdateAudioFFTCoeffs(uint32_t ChannelIndex);
    void SetAudioMuteBits(uint32_t ChannelIndex,uint32_t Bits);
    void ClearAudioMuteBits(uint32_t ChannelIndex,uint32_t Bits);
    void StopAudio(uint32_t ChannelIndex);

    void UpdateAGC(uint32_t ChannelIndex);

    void StartDDC1(void);
    void StartDDC2(uint32_t ChannelIndex);
    void StartAudio(uint32_t ChannelIndex);
    void UpdateAudioVolumeCoeff(uint32_t ChannelIndex);

    static uint32_t GetSamplesPerADCSnapshot(uint32_t FFTSize);

    void Start(void);
    void Stop(void);

    inline static TRange GetRange(uint64_t FrontEndFrequency)
    {
        return FrontEndFrequency==0?Range_1:Range_2;
    }
};

#endif
