#include "example.h"
#include <stdio.h>
#include <math.h>
#include <dlfcn.h>
#include <inttypes.h>

#define AUDIO_SAMPLE_RATE 48000

TG69DDC::TG69DDC(void)
{
    m_Device=NULL;
    m_API=NULL;
    m_PaSimple=NULL;
}

TG69DDC::~TG69DDC(void)
{
    Close();
    UnloadAPI();
}

bool TG69DDC::LoadAPI(void)
{
    if(m_API)
    {
        return true;
    }

    m_API=dlopen("libg69ddcapi.so",RTLD_LAZY);

    if(m_API==NULL)
    {
        printf("Failed to load libg69ddcapi.so. %s\n",dlerror());
        return false;
    }

    m_CreateInstance=(G69DDC_CREATE_INSTANCE)dlsym(m_API,"CreateInstance");

    if(m_CreateInstance==NULL)
    {
        printf("Failed to retrieve address of the CreateInstance function!\n");
        dlclose(m_API);
        m_API=NULL;
        return false;
    }

    return true;
}

void TG69DDC::UnloadAPI(void)
{
    //release device object interface before the API is freed
    if(m_Device)
    {
        m_Device->Free();
        m_Device=NULL;
    }

    if(m_API)
    {
        dlclose(m_API);
        m_API=NULL;
    }
}

bool TG69DDC::Open(const char *DevicePath)
{
 G69DDC_DDC_INFO DDCInfo;
 uint32_t DDCTypeIndex;
 uint64_t Frequency; 
 uint64_t FrontEndFrequency;
 int32_t DDC1Frequency;
 int32_t DDC2Frequency;
 int32_t DemodulatorFrequency;
 
    if(m_Device)
    {
        //close it if it is already open, if it is not open, the Close method does nothing
        m_Device->Close();
    }
    else
    {
        //device object has not been created yet
        if(!m_CreateInstance(G69DDC_CLASS_ID_DEVICE,(void**)&m_Device))
        {
            printf("Failed to create device object!\n");
            return false;
        }
    }

    if(!m_Device->Open(DevicePath))
    {
        printf("Failed to open device!\n");
        return false;
    }

    //create DDC1 channel
    m_Device->CreateDDC1(&m_DDC1Id);
    
    //create DDC2 channel
    m_Device->CreateDDC2(m_DDC1Id,&m_DDC2Id);
    
    //power on
    m_Device->SetPower(1);

    //set DDC type of DDC1 to 1 (DDC Bandwidth=24000, SampleRate=32kHz, BitsPerSample=32)
    //number of available DDC types of DDC1 can be obtained using GetDDCCount method
    m_Device->SetDDC(m_DDC1Id,1);

    //retrieve current DDC type of the DDC1
    m_Device->GetDDC(m_DDC1Id,&DDCTypeIndex,&DDCInfo);    
    printf("DDC type of DDC1: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);


    //set DDC type of DDC2 to 1 (DDC Bandwidth=24000, SampleRate=32kHz, BitsPerSample=32)    
    m_Device->SetDDC(m_DDC2Id,1);

    //retrieve current DDC type of the DDC2
    m_Device->GetDDC(m_DDC2Id,&DDCTypeIndex,&DDCInfo);    
    printf("DDC type of DDC2: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);

    //register callback object to this, the TG69DDC class implements methods of the IG69DDCDeviceCallback callback interface 
    m_Device->SetCallback(this);

    //tune to an AM station frequency
    m_Device->SetFrequency(m_DDC2Id,702000);
    //it is possible to tune DDC1, DDC2 and demodulator directly using
    //SetDDCFrequency and SetDemodulatorFrequency
    //m_Device->SetFrequency(DDC2Id,702000) can be replaced by the following:
    //m_Device->SetFrontEndFrequency(0);
    //m_Device->SetDDCFrequency(DDC1Id,702000);
    //m_Device->SetDDCFrequency(DDC2Id,0);
    //m_Device->SetDemodulatorFrequency(DDC2Id,0);
    //final absolute frequency of the demodulator is 702000+0+0=702000

    //get absolute frequency where the demodulator of channel 0 is tuned
    m_Device->GetFrequency(m_DDC2Id,&Frequency);    

    //get front-end frequency
    m_Device->GetFrontEndFrequency(&FrontEndFrequency);
    
    //get relative DDC1 frequency
    m_Device->GetDDCFrequency(m_DDC1Id,&DDC1Frequency);

    //get relative DDC2 frequency
    m_Device->GetDDCFrequency(m_DDC2Id,&DDC2Frequency);

    //get relative demodulator frequency
    m_Device->GetDemodulatorFrequency(m_DDC2Id,&DemodulatorFrequency);

    printf("Listening to (demodulator absolute frequency): %" PRIu64 " Hz\n",Frequency);
    printf("Front-end frequency: %" PRIu64 " Hz (range %d)\n",FrontEndFrequency,FrontEndFrequency==0?1:2);
    printf("DDC1 frequency (relative): %d Hz\n",DDC1Frequency);
    printf("DDC2 frequency (relative): %d Hz\n",DDC2Frequency);
    printf("Demodulator frequency (relative): %d Hz\n",DemodulatorFrequency);


    //set AGC params, attack=0.02s, decay=0.5s, ref. level=-15dB
    m_Device->SetAGCParams(m_DDC2Id,0.020,0.500,-15.0);
    
    //enable AGC
    m_Device->SetAGC(m_DDC2Id,1);

    //set AM mode
    m_Device->SetDemodulatorMode(m_DDC2Id,G69DDC_MODE_AM);

    //set demodulator filter bandwidth to 8kHz
    m_Device->SetDemodulatorFilterBandwidth(m_DDC2Id,8000);

    return true;
}

void TG69DDC::StartStreaming(void)
{
 uint32_t SamplesPerBuffer;
 G69DDC_DDC_INFO DDCInfo;

    if(m_Device)
    {
        //start DDC1 streaming using ~50ms long buffers
        m_Device->GetDDC(m_DDC1Id,NULL,&DDCInfo);
        SamplesPerBuffer=50*DDCInfo.SampleRate/1000;    //calculte number of I/Q sample-sets per 50ms buffer
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartDDC(m_DDC1Id,SamplesPerBuffer);    //start DDC1 streaming

        //start DDC2 streaming using ~50ms long buffers
        m_Device->GetDDC(m_DDC2Id,NULL,&DDCInfo);       
        SamplesPerBuffer=50*DDCInfo.SampleRate/1000;    //calculte number of I/Q sample-sets per 50ms buffer
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartDDC(m_DDC2Id,SamplesPerBuffer);        //start DDC2 streaming

        //start audio streaming using ~50ms long buffers
        //default audio sampling rate is 48000 Hz
        SamplesPerBuffer=50*AUDIO_SAMPLE_RATE/1000;     //calculte number of stereo audio sample-sets per 50ms buffer        
        SamplesPerBuffer=(SamplesPerBuffer+63)&(~63);   //round up to 64
        m_Device->StartAudio(m_DDC2Id,SamplesPerBuffer);       //start audio streaming
    }
}

void TG69DDC::Close(void)
{
    if(m_Device)
    {
        //stop streaming if it is running
        StopStreaming();

        //close wave out if it is open
        CloseWaveOut();

        //close the receiver if it is open
        m_Device->Close();
    }
}

void TG69DDC::StopStreaming(void)
{
    if(m_Device)
    {
        //stop audio, DDC2 and DDC1 streaming for channel 0
        m_Device->StopAudio(m_DDC2Id);
        m_Device->StopDDC(m_DDC2Id);
        m_Device->StopDDC(m_DDC1Id);
    }
}

void TG69DDC::OpenWaveOut(void)
{ 
 pa_sample_spec SampleFormat;
 int Error;
 
    if(m_PaSimple || m_Device==NULL)
    {
        return;
    }
           
    //initialize pulse audio
    SampleFormat.channels=2;
    SampleFormat.rate=AUDIO_SAMPLE_RATE;
    SampleFormat.format=PA_SAMPLE_FLOAT32LE;
    
    m_PaSimple=pa_simple_new(NULL,"G69DDC SDK EXAMPLE",PA_STREAM_PLAYBACK,NULL,"Audio0",&SampleFormat,NULL,NULL,&Error);
    
    if(m_PaSimple==NULL)
    {
        printf("Failed to initialize pulse audio.\n");
    }
}

void TG69DDC::CloseWaveOut(void)
{
    if(m_PaSimple)
    {
        pa_simple_free(m_PaSimple);
        m_PaSimple=NULL;
    }
}


void TG69DDC::G69DDC_DDC1StreamCallback(IG69DDCDevice *Device,uint32_t DDC1ChannelID,const void *Buffer,uint32_t Count,uint32_t BitsPerSample)
{
    //it does nothing in this example
}

int TG69DDC::G69DDC_DDC1PlaybackStreamCallback(IG69DDCDevice *Device,uint32_t DDC1ChannelId,void *Buffer,uint32_t Count,uint32_t BitsPerSample)
{
    //it does nothing in this example
    return 0;
}

void TG69DDC::G69DDC_DDC2StreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,const float *Buffer,uint32_t Count)
{
    //it does nothing in this example
}

void TG69DDC::G69DDC_DDC2PreprocessedStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,const float *Buffer,uint32_t Count,float SlevelPeak,float SlevelRMS)
{
 double Slevel_dBm;

    if(DDC2ChannelId==m_DDC2Id)
    {
        //convert slevel RMS in Volts to dBm
        Slevel_dBm=10.0*log10(SlevelRMS*SlevelRMS*(1000.0/50.0));

        printf("\rSlevel: RMS[V]=%9.7f V, RMS[dBm]=%6.1f dBm, Peak[V]=%9.7f V",SlevelRMS,Slevel_dBm,SlevelPeak);
        fflush(stdout);
    }
}

void TG69DDC::G69DDC_ADCSnapshotCallback(IG69DDCDevice *Device,const int16_t *Buffer,uint32_t Count,uint32_t CenterFrequency,uint16_t ADCLevel)
{
    //it does nothing in this example
}


void TG69DDC::G69DDC_AudioStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,uint32_t Stage,const float *Buffer,uint32_t Count)
{
 int Error;

    if(DDC2ChannelId==m_DDC2Id && Stage==2 && m_PaSimple!=NULL)
    {
        //we want to play only audio buffers of type 2 which means final audio samples (audio gain, audio filter and volume are applied)
        pa_simple_write(m_PaSimple,Buffer,Count*2*sizeof(float),&Error);
    } 
}

int TG69DDC::G69DDC_AudioPlaybackStreamCallback(IG69DDCDevice *Device,uint32_t DDC2ChannelId,float *Buffer,uint32_t Count)
{
    //it does nothing in this example
    return 0;
}

int TG69DDC::G69DDC_SweeperCallback(IG69DDCDevice *Device,const short *Buffer,uint32_t Count,uint32_t CenterFrequency,uint64_t FrontEndFrequency,uint32_t Repeat)
{
    //it does nothing in this example
    return 0;
}

int main(void)
{
 TG69DDC g69ddc;

    //load G69DDC API
    if(!g69ddc.LoadAPI())
    {
        return -1;
    }

    //open the first availbe G69DDC receiver and setup it
    if(!g69ddc.OpenFirst())
    {
        return -1;
    }
    
    g69ddc.OpenWaveOut();
    
    printf("\nPress ENTER to exit\n\n");

    //start streaming
    g69ddc.StartStreaming();

    //wait for ENTER
    getchar();

    //stop streaming
    g69ddc.StopStreaming();    

    //close the G69DDC receiver
    g69ddc.Close();
    
    g69ddc.CloseWaveOut();

    //unload G69DDC API
    g69ddc.UnloadAPI();
    
    return 0;
}
